@@ -12,6 +12,7 @@ import numpy as np
 
 CubicHermiteInterp = deutil.interpolation.CubicHermiteInterp
 root_finder        = deutil.optimizer.brentsrootvec
+root_polisher      = deutil.optimizer.newtonraphson
 
 __all__ = [
     'DiffRHS',
@@ -19,7 +20,7 @@ __all__ = [
     'OdeSystem'
 ]
 
-StateTuple = collections.namedtuple('StateTuple', ['t', 'y'])
+StateTuple = collections.namedtuple('StateTuple', ['t', 'y', 'event'])
 
 ##### Code adapted from https://github.com/scipy/scipy/blob/v1.3.2/scipy/integrate/_ivp/ivp.py#L28 #####
 def prepare_events(events):
@@ -30,21 +31,22 @@ def prepare_events(events):
     if events is not None:
         is_terminal = D.zeros(len(events), dtype=bool)
         direction   = D.zeros(len(events), dtype=D.int64)
+        last_occurence = D.zeros(len(events), dtype=D.int64) - 1
         for i, event in enumerate(events):
             if hasattr(event, "is_terminal"):
                 is_terminal[i] = bool(event.is_terminal)
-
             if hasattr(event, "direction"):
                 direction[i] = event.direction
     else:
         is_terminal = None
         direction   = None
+        last_occurence = None
 
-    return events, is_terminal, direction
+    return events, is_terminal, direction, last_occurence
 
 #####
 
-def handle_events(sol, events, consts, direction, is_terminal):
+def handle_events(sol_tuple, events, consts, direction, is_terminal):
     """Helper function to handle events.
     Parameters
     ----------
@@ -70,28 +72,46 @@ def handle_events(sol, events, consts, direction, is_terminal):
     terminate : bool
         Whether a terminal event occurred.
     """
+    sol, t_prev, t_next = sol_tuple
     ev_f = [(lambda event: lambda t: event(t, sol(t), **consts))(ev) for ev in events]
-    t_prev, t_next = sol.t0, sol.t1
     
     roots, success = root_finder(
         ev_f,
         [t_prev, t_next],
-        tol=D.epsilon()
+        tol=None,
+        verbose=False
     )
     
     roots = D.asarray(roots)
     
-    int_dir = D.sign(t_next - t_prev)
+    g     = [ev_f[idx](t_root - (t_next - t_prev)*D.epsilon()**0.5) for idx, t_root in enumerate(roots)]
+    g_cen = [ev_f[idx](t_root) for idx, t_root in enumerate(roots)]
+    g_new = [ev_f[idx](t_root + (t_next - t_prev)*D.epsilon()**0.5) for idx, t_root in enumerate(roots)]
     
-    g     = [events[idx]((1.9*t_root + 0.1*t_prev)/2, sol((1.9*t_root + 0.1*t_prev)/2), **consts) for idx, t_root in enumerate(roots)]
-    g_new = [events[idx]((1.9*t_root + 0.1*t_next)/2, sol((1.9*t_root + 0.1*t_next)/2), **consts) for idx, t_root in enumerate(roots)]
+    g     = D.to_float(D.stack(g))
+    g_cen = D.to_float(D.stack(g_cen))
+    g_new = D.to_float(D.stack(g_new))
     
-    g     = D.stack(g)
-    g_new = D.stack(g_new)
+    if D.backend() == 'torch':
+        direction = direction.to(g.device)
     
-    up     = (g <= 0) & (g_new >= 0)
-    down   = (g >= 0) & (g_new <= 0)
+    up     = ((g <= 0) & (g_new >= 0)) | ((g <= 0) & (g_cen >= 0)) | ((g_cen <= 0) & (g_new >= 0))
+    down   = ((g >= 0) & (g_new <= 0)) | ((g >= 0) & (g_cen <= 0)) | ((g_cen >= 0) & (g_new <= 0))
+    
+    for receptive_field in [1.0, 2.0, 3.0]:
+        g     = [ev_f[idx](t_root - receptive_field * (t_next - t_prev)*D.epsilon()**0.75) for idx, t_root in enumerate(roots)]
+        g_new = [ev_f[idx](t_root + receptive_field * (t_next - t_prev)*D.epsilon()**0.75) for idx, t_root in enumerate(roots)]
+
+        g     = D.to_float(D.stack(g))
+        g_new = D.to_float(D.stack(g_new))
+
+        up     = up | (((g <= 0) & (g_new >= 0)) | ((g <= 0) & (g_cen >= 0)) | ((g_cen <= 0) & (g_new >= 0)))
+        down   = down | ((g >= 0) & (g_new <= 0)) | ((g >= 0) & (g_cen <= 0)) | ((g_cen >= 0) & (g_new <= 0))
+    
+    up   = success & up
+    down = success & down
     either = up | down
+    
     mask   = (up     & (direction > 0) |
               down   & (direction < 0) |
               either & (direction == 0))
@@ -102,20 +122,23 @@ def handle_events(sol, events, consts, direction, is_terminal):
         active_events = D.reshape(D.nonzero(mask)[0], (-1,))
         
     roots     = roots[active_events]
+    evs       = [events[idx] for idx in active_events]
     terminate = False
     
     if len(active_events) > 0:
-        order = D.argsort(D.sign(t_next - t_prev) * roots)
+        order         = D.argsort(D.sign(t_next - t_prev) * roots)
         active_events = active_events[order]
         roots         = roots[order]
+        evs           = [evs[idx] for idx in order]
         
         if D.any(is_terminal[active_events]):
             t             = D.nonzero(is_terminal[active_events])[0][0]
             active_events = active_events[:t + 1]
             roots         = roots[:t + 1]
+            evs           = evs[:t + 1]
             terminate     = True
             
-    return active_events, roots, terminate
+    return active_events, roots, terminate, evs
 
 class DenseOutput(object):
     """Dense Output class for storing the dense output from a numerical integration.
@@ -129,7 +152,9 @@ class DenseOutput(object):
     """
     def __init__(self, t_eval, y_interpolants):
         if t_eval is None and y_interpolants is None:
-            self.t_eval         = [0.0]
+            self.t_eval         = [D.array(0.0)]
+            self.__t_eval_arr   = D.stack(self.t_eval)
+            self.__t_eval_arr_stale = False
             self.y_interpolants = []
         else:
             if t_eval is None or y_interpolants is None:
@@ -137,39 +162,71 @@ class DenseOutput(object):
             elif len(t_eval) != len(y_interpolants) + 1:
                 raise ValueError("The number of evaluation times and interpolants must be equal!")
             else:
-                self.t_eval         = list(t_eval)
+                self.t_eval         = [D.asarray(t) for t in t_eval]
+                self.__t_eval_arr   = D.stack(self.t_eval)
+                self.__t_eval_arr_stale = False
                 self.y_interpolants = y_interpolants
         
+    @property
+    def t_eval_arr(self):
+        if self.__t_eval_arr_stale:
+            self.__t_eval_arr = D.stack(self.t_eval)
+            self.__t_eval_arr_stale = False
+        return self.__t_eval_arr
+        
+    def find_interval(self, t):
+        return min(deutil.search_bisection(self.t_eval, t), len(self.y_interpolants) - 1)
+        
+    def find_interval_vec(self, t):
+        out = deutil.search_bisection_vec(self.t_eval_arr, t)
+        out[out > len(self.y_interpolants) - 1] = len(self.y_interpolants) - 1
+        return out
+        
     def __call__(self, t):
         if len(D.shape(t)) > 0:
-            ret_vals = D.zeros_like(D.asarray(t))
-            flat_t   = D.reshape(D.asarray(t), (-1,))
-            _y_test  = self.y_interpolants[0](self.t_eval[0])
-            flat_y   = D.stack([D.empty_like(_y_test) for _ in range(len(flat_t))])
-            for idx, _t in enumerate(flat_t):
-                tidx = min(deutil.search_bisection(self.t_eval, _t), len(self.y_interpolants) - 1)
-                flat_y[idx] = self.y_interpolants[tidx](_t)
-            return D.reshape(flat_y, D.shape(t) + D.shape(flat_y)[1:])
+            __flat_t  = D.reshape(D.asarray(t), (-1,))
+            __indices = self.find_interval_vec(__flat_t)
+            y_vals   = D.stack([
+                self.y_interpolants[idx](_t) for idx,_t in zip(__indices, __flat_t)
+            ], axis=0)
+            return D.reshape(y_vals, D.shape(t) + D.shape(y_vals)[1:])
         else:
-            tidx = min(deutil.search_bisection(self.t_eval, t), len(self.y_interpolants) - 1)
+            tidx = self.find_interval(t)
             return self.y_interpolants[tidx](t)
     
     def add_interpolant(self, t, y_interp):
-        try:
-            y_interp(self.t_eval[-1])
-        except:
-            raise
-        try:
-            y_interp(t)
-        except:
-            raise
-        if (t - self.t_eval[-1]) < 0:
-            self.t_eval.insert(0, t)
-            self.y_interpolants.insert(0, y_interp)
+        if isinstance(t, list) and isinstance(y_interp, list):
+            assert (len(t) == len(y_interp))
+            for idx in range(len(t)):
+                self.add_interpolant(t[idx], y_interp[idx])
+        elif (isinstance(t, list) and not isinstance(y_interp, list)) or (not isinstance(t, list) and isinstance(y_interp, list)):
+            raise TypeError("Expected both t and y_interp to be lists, but got type(t)={}, type(y_interp)={}".format(type(t), type(y_interp)))
         else:
-            self.t_eval.append(t)
-            self.y_interpolants.append(y_interp)
+            try:
+                y_interp(self.t_eval[-1])
+            except:
+                raise
+            try:
+                y_interp(t)
+            except:
+                raise
+            if (t - self.t_eval[-1]) < 0:
+                self.t_eval.insert(0, D.asarray(t))
+                self.y_interpolants.insert(0, y_interp)
+            else:
+                self.t_eval.append(D.asarray(t))
+                self.y_interpolants.append(y_interp)
+            if D.backend() == 'torch':
+                self.t_eval = [i.to(D.asarray(t)) for i in self.t_eval]
+            self.__t_eval_arr_stale = True
     
+    def remove_interpolant(self, idx):
+        out = self.t_eval.pop(idx),   self.y_interpolants.pop(idx)
+        self.__t_eval_arr = D.stack(self.t_eval)
+        return out
+    
+    def __len__(self):
+        return len(self.t_eval) - 1
 
 class DiffRHS(object):
     """Differential Equation class. Designed to wrap around around a function for the right-hand side of an ordinary differential equation.
@@ -207,14 +264,24 @@ class DiffRHS(object):
                 except:
                     self.md_repr = str(self.rhs)
             self.equ_repr = str(self.rhs)
+        self.__jac_wrapped_rhs_order = None
         if hasattr(self.rhs, 'jac'):
             self.__jac = self.rhs.jac
             self.__jac_time = None
             self.__jac_is_wrapped_rhs = False
-        else:
-            self.__jac = deutil.JacobianWrapper(lambda y, **kwargs: self(0.0, y, **kwargs), flat=True)
+        elif D.backend() != 'torch':
+            self.__jac_wrapped_rhs_order = 5
+            self.__jac = deutil.JacobianWrapper(lambda y, **kwargs: self(0.0, y, **kwargs), base_order=self.__jac_wrapped_rhs_order, flat=True)
             self.__jac_time = 0.0
             self.__jac_is_wrapped_rhs = True
+        else:
+            def __jac(t,y,*args,**kwargs):
+                y_in = y.clone().detach()
+                y_in.requires_grad = True
+                return D.jacobian(self(t,y_in,*args,**kwargs), y_in)
+            self.__jac = __jac
+            self.__jac_time = None
+            self.__jac_is_wrapped_rhs = False
         self.nfev = 0
         self.njev = 0
         
@@ -236,7 +303,7 @@ class DiffRHS(object):
         if self.__jac_is_wrapped_rhs:
             if t != self.__jac_time:
                 self.__jac_time = t
-                self.__jac = deutil.JacobianWrapper(lambda y, **kwargs: self(t, y, **kwargs), flat=True)
+                self.__jac = deutil.JacobianWrapper(lambda y, **kwargs: self(t, y, **kwargs), base_order=self.__jac_wrapped_rhs_order, flat=True)
             called_val = self.__jac(y, *args, **kwargs)
         else:
             called_val = self.__jac(t, y, *args, **kwargs)
@@ -260,9 +327,17 @@ class DiffRHS(object):
         estimate if a jacobian function was originally attached.
         """
         if not self.__jac_is_wrapped_rhs:
-            self.__jac                = deutil.JacobianWrapper(lambda y, **kwargs: self.rhs(0.0, y, **kwargs), flat=True)
+            if self.__jac_wrapped_rhs_order is None:
+                self.__jac_wrapped_rhs_order = 5
+            self.__jac                = deutil.JacobianWrapper(lambda y, **kwargs: self.rhs(0.0, y, **kwargs), base_order=self.__jac_wrapped_rhs_order, flat=True)
             self.__jac_time           = 0.0
             self.__jac_is_wrapped_rhs = True
+            
+    def set_jac_base_order(self, order):
+        if self.__jac_is_wrapped_rhs:
+            self.__jac_wrapped_rhs_order = order
+            self.__jac = deutil.JacobianWrapper(lambda y, **kwargs: self.rhs(0.0, y, **kwargs), base_order=self.__jac_wrapped_rhs_order, flat=True)
+            self.__jac_time = 0.0
     
     def __str__(self):
         return self.equ_repr
@@ -286,6 +361,19 @@ class DiffRHS(object):
             __new_diff_rhs.hook_jacobian_call(copy.deepcopy(self.__jac, memo))
         return __new_diff_rhs
 
+    __base_attributes = ["rhs", "equ_repr", "md_repr", "nfev", "njev", "__jac", "__jac_is_wrapped_rhs", "__jac_wrapped_rhs_order", "__jac_time"]
+    
+    def __getattr__(self, name):
+        return getattr(self.rhs, name)
+    
+    def __setattr__(self, name, val):
+        if name.replace("_DiffRHS", "") in self.__base_attributes:
+            self.__dict__[name] = val
+        elif name == "jac":
+            self.hook_jacobian_call(val)
+        else:
+            self.rhs.__dict__[name] = val
+
 def rhs_prettifier(equ_repr=None, md_repr=None):
     def rhs_wrapper(rhs):
         return DiffRHS(rhs, equ_repr, md_repr)
@@ -361,19 +449,26 @@ class OdeSystem(object):
             
         self.staggered_mask = None
         self.__dense_output = dense_output
-        self.int_status     = 0
-        self.success        = False
-        self.sol            = None
+        self.__int_status   = 0
+        self.__sol          = DenseOutput([self.t0], [])
             
         self.__move_to_device()
         self.__allocate_soln_space(self.__alloc_space_steps(self.tf))
         self.__fix_dt_dir(self.tf, self.t0)
         self.__events       = []
-        self.initialise_integrator()
+        self.initialise_integrator(preserve_states=False)
         
+    @property
+    def sol(self):
         if self.__dense_output:
-            self.sol = DenseOutput([self.t0], [])
-
+            return self.__sol
+        else:
+            return None
+        
+    @property
+    def success(self):
+        return self.__int_status == 1 or self.__int_status == 2
+            
     @property
     def events(self):
         """A tuple of (time, state) tuples at which each event occurs.
@@ -387,10 +482,7 @@ class OdeSystem(object):
         (StateTuple(t=..., y=...), StateTuple(t=..., y=...), StateTuple(t=..., y=...), ...)
         
         """
-        if self.__events:
-            return tuple(self.__events)
-        else:
-            return
+        return tuple(self.__events)
     
     @property
     def y(self):
@@ -627,8 +719,8 @@ class OdeSystem(object):
         """Returns the current time of the ODE system"""
         return self.t[self.counter]
 
-    def initialise_integrator(self):
-        integrator_kwargs = dict(dtype=self.y[0].dtype, device=self.device)
+    def initialise_integrator(self, preserve_states=False):
+        integrator_kwargs = dict(dtype=self.y[-1].dtype, device=self.device)
         
         integrator_kwargs['atol'] = self.atol
         integrator_kwargs['rtol'] = self.rtol
@@ -636,7 +728,19 @@ class OdeSystem(object):
         if self.method.__symplectic__ and not self.method.__implicit__:
             integrator_kwargs['staggered_mask'] = self.staggered_mask
             
+        if self.integrator:
+            old_states_exist = True
+            old_dState      = self.integrator.dState
+            old_dTime       = self.integrator.dTime
+        else:
+            old_states_exist = False
         self.integrator = self.method(self.dim, **integrator_kwargs)
+        if old_states_exist and preserve_states:
+            self.integrator.dState = old_dState
+            self.integrator.dTime  = old_dTime
+            if D.backend() == 'torch':
+                self.integrator.dState      = self.integrator.dState.to(self.integrator.device)
+                self.integrator.dTime       = self.integrator.dTime.to(self.integrator.device)
         
     def __get_integrator_mask(self, staggered_mask):
         if staggered_mask is None and hasattr(self.integrator, "staggered_mask"):
@@ -653,7 +757,7 @@ class OdeSystem(object):
     def method(self, new_method):
         self.set_method(new_method, None)
     
-    def set_method(self, new_method, staggered_mask=None):
+    def set_method(self, new_method, staggered_mask=None, preserve_states=True):
         """Sets the method of integration.
 
         Parameters
@@ -674,9 +778,8 @@ class OdeSystem(object):
             If the string is not a valid integration scheme.
         """        
         self.staggered_mask = self.__get_integrator_mask(staggered_mask)
-        if self.int_status == 1:
-            deutil.warning("An integration was already run, the system will be reset")
-            self.reset()
+        if self.__int_status == 1:
+            deutil.warning("An integration was already run, this will not reset the system", category=RuntimeWarning)
         if new_method in integrators.available_methods():
             self.__method = integrators.available_methods(False)[new_method]
         elif issubclass(new_method, integrators.IntegratorTemplate):
@@ -686,19 +789,12 @@ class OdeSystem(object):
         else:
             raise ValueError("The method you selected does not exist in the list of available methods, \
                               call desolver.available_methods() to see what these are")
-        self.initialise_integrator()
+        self.initialise_integrator(preserve_states=preserve_states)
                 
     def get_step_interpolant(self):
-        "Computes the 3rd order Hermite polynomial interpolant over one step."
-        return CubicHermiteInterp(
-                    self.__t[self.counter-1], 
-                    self.__t[self.counter], 
-                    self.__y[self.counter-1], 
-                    self.__y[self.counter],
-                    self.equ_rhs(self.__t[self.counter-1], self.__y[self.counter-1], **self.constants),
-                    self.equ_rhs(self.__t[self.counter], self.__y[self.counter], **self.constants)
-                )
+        return self.integrator.dense_output()
     
+    @property
     def integration_status(self):
         """Returns the integration status as a human-readable string.
 
@@ -707,36 +803,35 @@ class OdeSystem(object):
         str
             String containing the integration status message.
         """
-        if self.int_status == 0:
+        if self.__int_status == 0:
             return "Integration has not been run."
-        elif self.int_status == 1:
-            return "Integration completed successfully."
-        elif self.int_status == 2:
-            return "Integration terminated upon finding a triggered event."
-        elif self.int_status == -1:
-            return "Recursion limit was reached during integration, "+\
-                   "this can be caused by the adaptive integrator being unable "+\
-                   "to find a suitable step size to achieve the rtol/atol "+\
-                   "requirements."
-        elif self.int_status == -2:
+        if isinstance(self.__int_status, KeyboardInterrupt):
             return "A KeyboardInterrupt exception was raised during integration."
-        elif self.int_status == -3:
-            return "A generic exception was raised during integration."
+        if isinstance(self.__int_status, etypes.FailedIntegration):
+            if self.__int_status.__cause__ is not None:
+                return "The integration failed with the following exception: {}\n\tCaused by {}".format(self.__int_status, self.__int_status.__cause__)
+            else:
+                return "The integration failed with the following exception: {}".format(self.__int_status)
+        if self.__int_status == 1:
+            return "Integration completed successfully."
+        if self.__int_status == 2:
+            return "Integration terminated upon finding a triggered event."
+        else:
+            return "It should not be possible to initialise the system with this status"
+        
 
     def reset(self):
         """Resets the system to the initial time."""
         self.counter      = 0
         self.__trim_soln_space()
-        self.sol          = None
+        self.__sol        = DenseOutput([self.t0], [])
         self.dt           = self.__dt0
         self.equ_rhs.nfev = 0
         self.__move_to_device()
-        self.int_status   = 0
-        if self.__dense_output:
-            self.sol = DenseOutput([self.t0], [])
+        self.__int_status   = 0
         if self.__events:
             self.__events = []
-        self.initialise_integrator()
+        self.initialise_integrator(preserve_states=False)
 
     def integrate(self, t=None, callback=None, eta=False, events=None):
         """Integrates the system to a specified time.
@@ -782,24 +877,24 @@ class OdeSystem(object):
         else:
             tf = self.tf
             
-        if D.abs(tf - self.t[-1]) < D.epsilon():
+        if D.abs(tf - self.__t[self.counter]) < D.epsilon():
             return
         steps  = 0
             
-        events, is_terminal, direction = prepare_events(events)
+        events, is_terminal, direction, last_occurence = prepare_events(events)
         
         if D.to_numpy(tf) == np.inf and not any(is_terminal):
-            deutil.warning("Specifying an indefinite integration time with no terminal events will lead to memory issues.")
+            deutil.warning("Specifying an indefinite integration time with no terminal events can lead to memory issues if no event terminates the integration.", category=RuntimeWarning)
         
-        self.__fix_dt_dir(tf, self.t[-1])
+        self.__fix_dt_dir(tf, self.__t[self.counter])
 
-        if D.abs(self.dt) > D.abs(tf - self.t[-1]):
-            self.dt = D.abs(tf - self.t[-1])*0.5
+        if D.abs(self.dt) > D.abs(tf - self.__t[self.counter]):
+            self.dt = D.abs(tf - self.__t[self.counter])*0.5
 
         total_steps = self.__alloc_space_steps(tf)
         
         if eta:
-            tqdm_progress_bar = tqdm(total=int((tf - self.__t[self.counter])/self.dt))
+            tqdm_progress_bar = tqdm(total=int((tf - self.__t[self.counter])/self.dt)+1)
         else:
             tqdm_progress_bar = None
             
@@ -810,14 +905,13 @@ class OdeSystem(object):
                 callback = []
             else:
                 callback = [callback]
-            
+        
         end_int = False
-        self.equ_rhs.nfev = 0 if self.int_status == 1 else self.equ_rhs.nfev
         cState  = D.zeros_like(self.__y[self.counter])
         cTime   = D.zeros_like(self.__t[self.counter])        
         self.__allocate_soln_space(total_steps)
         try:
-            while self.dt != 0 and D.abs(tf - self.__t[self.counter]) > 4 * D.epsilon() and not end_int:
+            while self.dt != 0 and D.abs(tf - self.__t[self.counter]) >= D.epsilon() and not end_int:
                 if D.abs(self.dt + self.__t[self.counter]) > D.abs(tf):
                     self.dt = (tf - self.__t[self.counter])
                 self.dt, (dTime, dState) = self.integrator(self.equ_rhs, self.__t[self.counter], self.__y[self.counter], self.constants, timestep=self.dt)
@@ -843,36 +937,43 @@ class OdeSystem(object):
                 self.counter += 1
 
                 if events is not None or self.__dense_output:
-                    tsol = self.get_step_interpolant()
-
+                    __pre_length = len(self.__sol)
+                    __t_interp, __y_interp = self.get_step_interpolant()
+                    self.__sol.add_interpolant(__t_interp, __y_interp)
+                    
                     if events is not None:
-                        active_events, roots, end_int = handle_events(tsol, events, self.constants, direction, is_terminal)
+                        next_time  = self.__t[self.counter]
+                        next_state = self.__y[self.counter]
+                        prev_time  = self.__t[self.counter - 1]
+                        prev_state = self.__y[self.counter - 1]
+                        self.counter -= 1
+                        
+                        sol_tuple = (self.__sol, prev_time, next_time)
+                        active_events, roots, end_int, evs = handle_events(sol_tuple, events, self.constants, direction, is_terminal)
 
                         if self.counter+len(roots)+1 >= len(self.__y):
                             total_steps = self.__alloc_space_steps(tf - dTime) + 1 + len(roots)
                             self.__allocate_soln_space(total_steps)
 
-                        prev_time  = self.__t[self.counter - 1]
-                        prev_state = self.__y[self.counter - 1]
-                        
-                        self.counter -= 1
-
-                        for root in roots:
-                            if root != self.__t[self.counter]:
-# #                                 _, (new_dTime, new_dState) = self.integrator(self.equ_rhs, self.__t[self.counter], self.__y[self.counter], self.constants, timestep=root - self.__t[self.counter])
-#                                 self.__t[self.counter+1] = root
-#                                 self.__y[self.counter+1] = tsol(root) # self.__y[self.counter] + new_dState
-#                                 self.__events.append(StateTuple(t=self.__t[self.counter+1], y=self.__y[self.counter+1]))
-#                                 self.counter += 1
-                                prev_dt = self.dt
-                                self.integrate(root)
-                                self.__events.append(StateTuple(t=self.__t[self.counter], y=self.__y[self.counter]))
-                                self.dt = prev_dt
-
+                        for ev_idx,(root,ev) in enumerate(zip(roots, evs)):
+                            if dTime >= 0:
+                                true_positive = (self.__t[self.counter] <= root) & (root <= prev_time + dTime)
+                            else:
+                                true_positive = (prev_time + dTime <= root) & (root <= self.__t[self.counter]) 
+                                
+#                             print(root, prev_time, prev_time + dTime, true_positive, ev(root, sol_tuple[0](root), **self.constants))
+                            if true_positive:
+                                ev_state = StateTuple(t=root, y=self.__sol(root), event=ev)
+                                if not self.__events or last_occurence[ev_idx] == -1:
+                                    last_occurence[ev_idx] = len(self.__events)
+                                    self.__events.append(ev_state)
+                                elif D.abs(ev_state.t - self.__events[last_occurence[ev_idx]].t) > D.epsilon()**0.7:
+                                    last_occurence[ev_idx] = len(self.__events)
+                                    self.__events.append(ev_state)
+                                    
                         if end_int:
-                            tsol            = self.get_step_interpolant()
-                            self.int_status = 2
-                            self.success    = True
+                            self.integrate(roots[-1])
+                            self.__int_status = 2
                         else:
                             if self.counter+len(roots)+1 >= len(self.__y):
                                 total_steps = self.__alloc_space_steps(tf - dTime) + 1 + len(roots)
@@ -880,43 +981,44 @@ class OdeSystem(object):
                             self.__t[self.counter+1] = prev_time  + dTime 
                             self.__y[self.counter+1] = prev_state + dState
                             self.counter += 1
-                    elif self.__dense_output:
-                        self.sol.add_interpolant(self.__t[self.counter], tsol)
-
+                            
+                    if not self.__dense_output:
+                        for _ in range(__pre_length-1):
+                            self.__sol.remove_interpolant(0)
+                
+                steps += 1
+                
+                for i in callback:
+                    i(self)
+                    
                 if tqdm_progress_bar is not None:
                     tqdm_progress_bar.total = tqdm_progress_bar.n
                     if D.to_numpy(tf) == np.inf:
                         tqdm_progress_bar.total = None
                     else:
-                        tqdm_progress_bar.total = tqdm_progress_bar.n + int((tf - self.__t[self.counter])/self.dt)
+                        tqdm_progress_bar.total = tqdm_progress_bar.n + int((tf - self.__t[self.counter])/self.dt) + 1
                     tqdm_progress_bar.desc  = "{:>10.2f} | {:.2f} | {:<10.2e}".format(self.__t[self.counter], tf, self.dt).ljust(8)
                     tqdm_progress_bar.update()
-
-                steps += 1
-
-                for i in callback:
-                    i(self)
                     
-        except KeyboardInterrupt:
-            self.int_status = -2
-            raise
-        except RecursionError:
-            self.int_status = -1
-            raise
-        except:
-            self.int_status = -3
-            raise
+        except KeyboardInterrupt as e:
+            self.__int_status = e
+            raise e
+        except Exception as e:
+            new_e = etypes.FailedIntegration("Failed to integrate system")
+            new_e.__cause__ = e
+            self.__int_status = new_e
+            raise new_e
+        else:
+            if self.__int_status != 2 and not isinstance(self.__int_status, (etypes.FailedIntegration, KeyboardInterrupt)):
+                self.__int_status = 1
         finally:
-            self.success    = True
-            if self.int_status != 2:
-                self.int_status = 1
-            self.__trim_soln_space()
             if eta:
                 tqdm_progress_bar.close()
+            self.__trim_soln_space()
 
     def __repr__(self):
         return "\n".join([
-            """{:>10}: {:<128}""".format("message",   self.integration_status()),
+            """{:>10}: {:<128}""".format("message",   self.integration_status),
             """{:>10}: {:<128}""".format("nfev",      str(self.nfev)),
             """{:>10}: {:<128}""".format("sol",       str(self.sol)),
             """{:>10}: {:<128}""".format("t0",        str(self.t0)),
@@ -944,7 +1046,7 @@ class OdeSystem(object):
 {:>10}: {:<128}  
 ```
 """.format(
-            "message", self.integration_status(), 
+            "message", self.integration_status,
             "nfev", str(self.nfev), 
             "sol", str(self.sol), 
             "t0", str(self.t0), 
@@ -976,7 +1078,7 @@ class OdeSystem(object):
             if index > self.counter:
                 raise IndexError("index {} out of bounds for integrations with {} steps".format(index, self.counter+1))
             else:
-                return StateTuple(t=self.t[index], y=self.y[index])
+                return StateTuple(t=self.t[index], y=self.y[index], event=None)
         elif isinstance(index, slice):
             if index.start is not None:
                 start_idx = deutil.search_bisection(self.t[:self.counter+1], index.start)
@@ -990,13 +1092,19 @@ class OdeSystem(object):
                 step      = index.step
             else:
                 step      = 1
-            return StateTuple(t=self.t[start_idx:end_idx:step], y=self.y[start_idx:end_idx:step])
+            return StateTuple(t=self.t[start_idx:end_idx:step], y=self.y[start_idx:end_idx:step], event=None)
         else:
             if self.__dense_output and self.sol is not None:
-                return StateTuple(t=index, y=self.sol(index))
+                return StateTuple(t=index, y=self.sol(index), event=None)
             else:
                 nearest_idx = deutil.search_bisection(self.__t, index)
-                return StateTuple(t=self.t[nearest_idx], y=self.y[nearest_idx])
+                if nearest_idx < self.counter:
+                    if D.abs(D.to_float(self.t[nearest_idx] - index)) < D.abs(D.to_float(self.t[nearest_idx+1]-index)):
+                        return StateTuple(t=self.t[nearest_idx], y=self.y[nearest_idx], event=None)
+                    else:
+                        return StateTuple(t=self.t[nearest_idx+1], y=self.y[nearest_idx+1], event=None)
+                else:
+                    return StateTuple(t=self.t[nearest_idx], y=self.y[nearest_idx], event=None)
             
     def __len__(self):
         return self.counter + 1
