@@ -5,7 +5,7 @@ from .. import backend as D
 
 class GaussLegendre4(ImplicitRungeKuttaIntegrator):
     order = 4
-    __adaptive__   = True
+    __adaptive__   = False
     __symplectic__ = False
     __alt_names__  = tuple()
     
@@ -24,7 +24,7 @@ class GaussLegendre4(ImplicitRungeKuttaIntegrator):
     
 class GaussLegendre6(ImplicitRungeKuttaIntegrator):
     order = 6
-    __adaptive__   = True
+    __adaptive__   = False
     __symplectic__ = False
     __alt_names__  = tuple()
     
@@ -44,7 +44,7 @@ class GaussLegendre6(ImplicitRungeKuttaIntegrator):
     
 class LobattoIIIA2(ImplicitRungeKuttaIntegrator):
     order = 2
-    __adaptive__   = True
+    __adaptive__   = False
     __symplectic__ = False
     __alt_names__  = tuple()
     
@@ -59,7 +59,7 @@ class LobattoIIIA2(ImplicitRungeKuttaIntegrator):
     
 class LobattoIIIA4(ImplicitRungeKuttaIntegrator):
     order = 4
-    __adaptive__   = True
+    __adaptive__   = False
     __symplectic__ = False
     __alt_names__  = tuple()
     
@@ -80,8 +80,8 @@ class LobattoIIIB2(ImplicitRungeKuttaIntegrator):
     __alt_names__  = tuple()
     
     tableau = numpy.array(
-        [[0.0,   0.5,  -0.5],
-         [1.0,   0.5,   0.5]], dtype=numpy.float64
+        [[0.5,   0.5,  0.0],
+         [0.5,   0.5,  0.0]], dtype=numpy.float64
     )
 
     final_state = numpy.array(
@@ -90,7 +90,7 @@ class LobattoIIIB2(ImplicitRungeKuttaIntegrator):
     
 class LobattoIIIB4(ImplicitRungeKuttaIntegrator):
     order = 4
-    __adaptive__   = True
+    __adaptive__   = False
     __symplectic__ = False
     __alt_names__  = tuple()
     
@@ -111,8 +111,8 @@ class LobattoIIIC2(ImplicitRungeKuttaIntegrator):
     __alt_names__  = tuple()
     
     tableau = numpy.array(
-        [[0.5,   0.5,   0  ],
-         [0.5,   0.5,   0  ]], dtype=numpy.float64
+        [[0.0,   0.5,   -0.5],
+         [1.0,   0.5,    0.5]], dtype=numpy.float64
     )
 
     final_state = numpy.array(
@@ -143,7 +143,7 @@ class BackwardEuler(ImplicitRungeKuttaIntegrator):
     __alt_names__  = tuple()
     
     tableau = numpy.array(
-        [[1.0, 0.0]], dtype=numpy.float64
+        [[1.0, 1.0]], dtype=numpy.float64
     )
 
     final_state = numpy.array(
@@ -198,7 +198,7 @@ class DIRK3LStable(ImplicitRungeKuttaIntegrator):
     
 class RadauIA3(ImplicitRungeKuttaIntegrator):
     order = 3
-    __adaptive__   = True
+    __adaptive__   = False
     __symplectic__ = False
     __alt_names__  = tuple()
     
@@ -214,7 +214,7 @@ class RadauIA3(ImplicitRungeKuttaIntegrator):
     
 class RadauIA5(ImplicitRungeKuttaIntegrator):
     order = 5
-    __adaptive__   = True
+    __adaptive__   = False
     __symplectic__ = False
     __alt_names__  = tuple()
     
@@ -235,7 +235,7 @@ class RadauIA5(ImplicitRungeKuttaIntegrator):
     
 class RadauIIA3(ImplicitRungeKuttaIntegrator):
     order = 3
-    __adaptive__   = True
+    __adaptive__   = False
     __symplectic__ = False
     __alt_names__  = tuple()
     
