@@ -46,10 +46,10 @@ class RichardsonIntegratorTemplate(IntegratorTemplate):
 
     def __init__(self):
         raise NotImplementedError("Do not initialise this class directly!")
-        
-    def dense_output(self):
-        raise NotImplementedError("Do not use this class directly! How did you initialise it??")
 
+    def dense_output(self):
+        return self.__interpolant_times, self.__interpolants
+        
     def adaptive_richardson(self, rhs, t, y, constants, timestep):
         dt0, (dt_z, dy_z) = self.step(0, rhs, t, y, timestep, constants, 1)
         if dt_z < timestep:
@@ -79,10 +79,15 @@ class RichardsonIntegratorTemplate(IntegratorTemplate):
     def step(self, int_num, rhs, initial_time, initial_state, timestep, constants, num_intervals):
         dt_now, dstate_now = 0.0, 0.0
         dtstep = timestep / num_intervals
+        self.__interpolants = []
+        self.__interpolant_times = []
         for interval in range(num_intervals):
             dt, (dt_z, dy_z) = self.basis_integrators[int_num](rhs, initial_time + dt_now, initial_state + dstate_now, constants, dtstep)
             dt_now     = dt_now     + dt_z
             dstate_now = dstate_now + dy_z
+            __interp_t, __interp = self.basis_integrators[int_num].dense_output()
+            self.__interpolant_times.append(__interp_t)
+            self.__interpolants.append(__interp)
         return dtstep, (dt_now, dstate_now)
 
     def forward(self, rhs, initial_time, initial_state, constants, timestep):
@@ -93,18 +98,21 @@ class RichardsonIntegratorTemplate(IntegratorTemplate):
         
         new_timestep, redo_step = self.update_timestep(initial_state, self.dState, diff, initial_time, dt_z, tol=0.5 if self.__implicit__ else 0.9)
         if self.__symplectic__:
-            timestep = dt0
-            if new_timestep < timestep:
-                while new_timestep < timestep:
-                    timestep /= 2.0
+            timestep      = dt0
+            next_timestep = D.copy(dt0)
+            if (0.8*new_timestep+0.2*timestep) < next_timestep:
+                while new_timestep < next_timestep:
+                    next_timestep /= 2.0
             else:
-                while new_timestep > 2*timestep:
-                    timestep *= 2.0
+                while (0.8*new_timestep+0.2*timestep) > 2*next_timestep:
+                    next_timestep *= 2.0
                 redo_step = False
         else:
-            timestep = new_timestep
+            next_timestep = new_timestep
         if redo_step:
-            timestep, (self.dTime, self.dState) = self(rhs, initial_time, initial_state, constants, timestep)
+            timestep, (self.dTime, self.dState) = self(rhs, initial_time, initial_state, constants, next_timestep)
+        else:
+            timestep = next_timestep
             
         return timestep, (self.dTime, self.dState)
 
