@@ -13,30 +13,7 @@ __all__ = [
 class ExplicitIntegrator(IntegratorTemplate):
     __implicit__ = False
     
-    def __init__(self, sys_dim, aux_shape, adaptive, dtype=None, rtol=None, atol=None):
-        self.dim        = sys_dim
-        self.rtol       = rtol if rtol is not None else 32 * D.epsilon()
-        self.atol       = atol if atol is not None else 32 * D.epsilon()
-        self.adaptive   = adaptive
-        self.dtype      = dtype
-        self.aux        = D.zeros((*aux_shape, *self.dim), dtype=self.dtype)
-        self.dState     = D.zeros(self.dim, dtype=self.dtype)
-        self.dTime      = D.zeros(tuple(), dtype=self.dtype)
-        
-    def dense_output(self, rhs, initial_time, initial_state):
-        return CubicHermiteInterp(
-            initial_time, 
-            initial_time + self.dTime, 
-            initial_state, 
-            initial_state + self.dState,
-            rhs(initial_time, initial_state),
-            rhs(initial_time + self.dTime, initial_state + self.dState)
-        )
-    
-class ImplicitIntegrator(IntegratorTemplate):
-    __implicit__ = True
-    
-    def __init__(self, sys_dim, aux_shape, adaptive, dtype=None, rtol=None, atol=None):
+    def __init__(self, sys_dim, aux_shape, adaptive, dtype=None, rtol=None, atol=None, device=None):
         self.dim        = sys_dim
         self.numel      = 1
         for i in self.dim:
@@ -48,16 +25,32 @@ class ImplicitIntegrator(IntegratorTemplate):
         self.aux        = D.zeros((*aux_shape, *self.dim), dtype=self.dtype)
         self.dState     = D.zeros(self.dim, dtype=self.dtype)
         self.dTime      = D.zeros(tuple(), dtype=self.dtype)
-        
-    def dense_output(self, rhs, initial_time, initial_state):
-        return CubicHermiteInterp(
-            initial_time, 
-            initial_time + self.dTime, 
-            initial_state, 
-            initial_state + self.dState,
-            rhs(initial_time, initial_state),
-            rhs(initial_time + self.dTime, initial_state + self.dState)
-        )
+        self.device     = device
+        self.initial_state = None
+        self.initial_rhs   = None
+        self.initial_time  = None
+        self.final_rhs     = None
+    
+class ImplicitIntegrator(IntegratorTemplate):
+    __implicit__ = True
+    
+    def __init__(self, sys_dim, aux_shape, adaptive, dtype=None, rtol=None, atol=None, device=None):
+        self.dim        = sys_dim
+        self.numel      = 1
+        for i in self.dim:
+            self.numel *= int(i)
+        self.rtol       = rtol if rtol is not None else 32 * D.epsilon()
+        self.atol       = atol if atol is not None else 32 * D.epsilon()
+        self.adaptive   = adaptive
+        self.dtype      = dtype
+        self.aux        = D.zeros((*aux_shape, *self.dim), dtype=self.dtype)
+        self.dState     = D.zeros(self.dim, dtype=self.dtype)
+        self.dTime      = D.zeros(tuple(), dtype=self.dtype)
+        self.device     = device
+        self.initial_state = None
+        self.initial_rhs   = None
+        self.initial_time  = None
+        self.final_rhs     = None
 
 class ExplicitRungeKuttaIntegrator(ExplicitIntegrator):
     """
@@ -95,33 +88,59 @@ class ExplicitRungeKuttaIntegrator(ExplicitIntegrator):
     __symplectic__ = False
 
     def __init__(self, sys_dim, dtype=None, rtol=None, atol=None, device=None):
-        super().__init__(sys_dim, (D.shape(self.tableau)[0],), D.shape(self.final_state)[0] == 2, dtype, rtol, atol)
+        super().__init__(sys_dim, (D.shape(self.tableau)[0],), D.shape(self.final_state)[0] == 2, dtype, rtol, atol, device)
         if dtype is None:
             self.tableau     = D.array(self.tableau)
             self.final_state = D.array(self.final_state)
         else:
             self.tableau     = D.to_type(self.tableau, dtype)
             self.final_state = D.to_type(self.final_state, dtype)
+            
+        self.__fsal = bool(D.all(self.tableau[-1, 1:] == self.final_state[0][1:]))
+        self.fsal   = self.__fsal
         
         if D.backend() == 'torch':
-            self.aux         = self.aux.to(device)
-            self.dState      = self.dState.to(device)
-            self.dTime       = self.dTime.to(device)
-            self.tableau     = self.tableau.to(device)
-            self.final_state = self.final_state.to(device)
+            self.aux         = self.aux.to(self.device)
+            self.dState      = self.dState.to(self.device)
+            self.dTime       = self.dTime.to(self.device)
+            self.tableau     = self.tableau.to(self.device)
+            self.final_state = self.final_state.to(self.device)
             
     def forward(self, rhs, initial_time, initial_state, constants, timestep):
         if self.tableau is None:
             raise NotImplementedError("In order to use the fixed step integrator, subclass this class and populate the butcher tableau")
         else:
             tableau_idx_expand = tuple([slice(1, None, None)] + [None] * (self.aux.ndim - 1))
+            self.initial_state = D.copy(initial_state)
+            self.initial_time  = D.copy(initial_time)
+            self.initial_rhs   = None
+            if not self.__fsal:
+                self.final_rhs = None
 
             for stage in range(D.shape(self.aux)[0]):
-                current_state   = initial_state    + D.sum(timestep * self.tableau[stage][tableau_idx_expand] * self.aux, axis=0)
-                self.aux[stage] = rhs(initial_time + self.tableau[stage, 0]*timestep, current_state, **constants)
+                current_state   = initial_state    + timestep * D.sum(self.tableau[stage][tableau_idx_expand] * self.aux, axis=0)
+                if self.__fsal and self.final_rhs is not None and stage == 0:
+                    self.aux[stage] = self.final_rhs
+                else:
+                    self.aux[stage] = rhs(initial_time + self.tableau[stage, 0]*timestep, current_state, **constants)
+                
+                if stage == 0 and D.sum(self.tableau[0]) == 0.0:
+                    self.initial_rhs = self.aux[0]
+                elif self.__fsal and stage == D.shape(self.aux)[0] - 1:
+                    self.final_rhs   = self.aux[stage]
+            
+            if self.__fsal:
+                self.dState = self.aux[-1] / timestep
+            else:
+                self.dState = timestep * D.sum(self.final_state[0][tableau_idx_expand] * self.aux, axis=0)
                 
-            self.dState = timestep * D.sum(self.final_state[0][tableau_idx_expand] * self.aux, axis=0)
             self.dTime  = D.copy(timestep)
+                
+            if self.initial_rhs is None:
+                self.initial_rhs = rhs(initial_time, initial_state, **constants)
+                
+            if self.final_rhs is None:
+                self.final_rhs = rhs(initial_time + self.dTime, initial_state + self.dState, **constants)
             
             if self.adaptive:
                 diff = timestep * self.get_error_estimate(tableau_idx_expand)
@@ -132,9 +151,20 @@ class ExplicitRungeKuttaIntegrator(ExplicitIntegrator):
             return timestep, (self.dTime, self.dState)
         
     def get_error_estimate(self, tableau_idx_expand):
-        return D.sum((self.final_state[1] - self.final_state[0])[tableau_idx_expand] * self.aux, axis=0)
+        return D.sum(self.final_state[1][tableau_idx_expand] * self.aux, axis=0) - D.sum(self.final_state[0][tableau_idx_expand] * self.aux, axis=0)
 
     __call__ = forward
+    
+    def dense_output(self):
+        return (self.initial_time + self.dTime, 
+                utilities.interpolation.CubicHermiteInterp(
+                    self.initial_time, 
+                    self.initial_time + self.dTime, 
+                    self.initial_state, 
+                    self.initial_state + self.dState,
+                    self.initial_rhs,
+                    self.final_rhs
+                ))
 
 class ExplicitSymplecticIntegrator(ExplicitIntegrator):
     """
@@ -169,7 +199,7 @@ class ExplicitSymplecticIntegrator(ExplicitIntegrator):
     __symplectic__ = True
 
     def __init__(self, sys_dim, dtype=None, staggered_mask=None, rtol=None, atol=None, device=None):
-        super().__init__(sys_dim, (D.shape(self.tableau)[0],), False, dtype, rtol, atol)
+        super().__init__(sys_dim, (D.shape(self.tableau)[0],), False, dtype, rtol, atol, device)
         if staggered_mask is None:
             staggered_mask      = D.arange(sys_dim[0]//2, sys_dim[0], dtype=D.int64)
             self.staggered_mask = D.zeros(sys_dim, dtype=D.bool)
@@ -186,11 +216,11 @@ class ExplicitSymplecticIntegrator(ExplicitIntegrator):
         self.nmsk = D.logical_not(self.staggered_mask)
         
         if D.backend() == 'torch':
-            self.dState      = self.dState.to(device)
-            self.dTime       = self.dTime.to(device)
-            self.tableau     = self.tableau.to(device)
-            self.msk  = self.msk.to(self.tableau)
-            self.nmsk = self.nmsk.to(self.tableau)
+            self.dState  = self.dState.to(self.device)
+            self.dTime   = self.dTime.to(self.device)
+            self.tableau = self.tableau.to(self.device)
+            self.msk     = self.msk.to(self.tableau)
+            self.nmsk    = self.nmsk.to(self.tableau)
 
     def forward(self, rhs, initial_time, initial_state, constants, timestep):
         if self.tableau is None:
@@ -200,20 +230,36 @@ class ExplicitSymplecticIntegrator(ExplicitIntegrator):
             nmsk = self.nmsk
 
             current_time  = D.copy(initial_time)
-            current_state = D.copy(initial_state)
-            self.dState   = D.zeros_like(current_state)
+            self.dState  *= 0.0
+            self.initial_state = D.copy(initial_state)
+            self.initial_time  = D.copy(initial_time)
+            self.initial_rhs   = None
+            self.final_rhs     = None
 
             for stage in range(D.shape(self.tableau)[0]):
-                aux          = rhs(current_time, initial_state + self.dState, **constants) * timestep
+                aux          = timestep * rhs(current_time, initial_state + self.dState, **constants)
                 current_time = current_time + timestep * self.tableau[stage, 0]
                 self.dState += aux * self.tableau[stage, 1] * msk + aux * self.tableau[stage, 2] * nmsk
                 
             self.dTime = D.copy(timestep)
+            self.initial_rhs = rhs(current_time, initial_state, **constants)
+            self.final_rhs   = rhs(initial_time + self.dTime, initial_state + self.dState, **constants)
             
             return self.dTime, (self.dTime, self.dState)
         
     __call__ = forward
     
+    def dense_output(self):
+        return (self.initial_time + self.dTime, 
+                utilities.interpolation.CubicHermiteInterp(
+                    self.initial_time, 
+                    self.initial_time + self.dTime, 
+                    self.initial_state, 
+                    self.initial_state + self.dState,
+                    self.initial_rhs,
+                    self.final_rhs
+                ))
+    
 class ImplicitRungeKuttaIntegrator(ImplicitIntegrator):
     """
     A base class for all implicit Runge-Kutta methods with arbitrary Butcher Tableau.
@@ -250,7 +296,7 @@ class ImplicitRungeKuttaIntegrator(ImplicitIntegrator):
     __symplectic__ = False
 
     def __init__(self, sys_dim, dtype=None, rtol=None, atol=None, device=None):
-        super().__init__(sys_dim, (D.shape(self.tableau)[0],), D.shape(self.final_state)[0] == 2, dtype, rtol, atol)
+        super().__init__(sys_dim, (D.shape(self.tableau)[0],), D.shape(self.final_state)[0] == 2, dtype, rtol, atol, device)
         if dtype is None:
             self.tableau     = D.array(self.tableau)
             self.final_state = D.array(self.final_state)
@@ -259,11 +305,11 @@ class ImplicitRungeKuttaIntegrator(ImplicitIntegrator):
             self.final_state = D.to_type(self.final_state, dtype)
         
         if D.backend() == 'torch':
-            self.dState      = self.dState.to(device)
-            self.dTime       = self.dTime.to(device)
-            self.aux         = self.aux.to(device)
-            self.tableau     = self.tableau.to(device)
-            self.final_state = self.final_state.to(device)        
+            self.dState      = self.dState.to(self.device)
+            self.dTime       = self.dTime.to(self.device)
+            self.aux         = self.aux.to(self.device)
+            self.tableau     = self.tableau.to(self.device)
+            self.final_state = self.final_state.to(self.device)
             
     def forward(self, rhs, initial_time, initial_state, constants, timestep):
         if self.tableau is None:
@@ -271,64 +317,70 @@ class ImplicitRungeKuttaIntegrator(ImplicitIntegrator):
         else:
             tableau_idx_expand = tuple([slice(1, None, None)] + [None] * (self.aux.ndim - 1))
 
+        self.initial_state = D.copy(initial_state)
+        self.initial_time  = D.copy(initial_time)
+        self.initial_rhs   = None
+        self.final_rhs     = None
         aux_shape = self.aux.shape
         
         def nfun(next_state):
+            nonlocal initial_time, initial_state, timestep, aux_shape
             __aux_states = D.reshape(next_state, aux_shape)
             __rhs_states = D.stack([
                 rhs(initial_time + tbl[0] * timestep, initial_state + timestep * D.sum(tbl[tableau_idx_expand] * __aux_states, axis=0), **constants) for tbl in self.tableau
             ])
             __states =  D.reshape(__aux_states - __rhs_states, (-1,))
-            if __states.dtype == object:
-                __states = D.to_float(__states)
             return __states
         
         def __nfun_jac(next_state):
-            nonlocal self
+            nonlocal self, initial_time, initial_state, timestep, aux_shape
             __aux_states = D.reshape(next_state, aux_shape)
             __step       = self.numel
-            __jac        = D.eye(self.tableau.shape[0]*__step)
+            if D.backend() == 'torch':
+                __jac        = D.eye(self.tableau.shape[0]*__step, device=__aux_states.device, dtype=__aux_states.dtype)
+            else:
+                __jac        = D.eye(self.tableau.shape[0]*__step)
             __prev_idx   = -1
             __rhs_jac = D.stack([
                 rhs.jac(initial_time + tbl[0] * timestep, initial_state + timestep * D.sum(tbl[tableau_idx_expand] * __aux_states, axis=0), **constants) for tbl in self.tableau
             ])
             for idx in range(0,__jac.shape[0],__step):
                 for jdx in range(0,__jac.shape[1],__step):
-                    __jac[idx:idx+__step, jdx:jdx+__step] -= timestep * self.tableau[idx//__step, 1+jdx//__step]*__rhs_jac[idx//__step]
+                    __jac[idx:idx+__step, jdx:jdx+__step] -= timestep * self.tableau[idx//__step, 1+jdx//__step]*__rhs_jac[idx//__step].reshape(__step, __step)
             if __jac.shape[0] == 1 and __jac.shape[1] == 1:
                 __jac = D.reshape(__jac, tuple())
-            if __jac.dtype == object:
-                __jac = D.to_float(__jac)
             return __jac
             
-        initial_guess = D.to_float((D.zeros_like(self.aux) + self.dState[None]))
-        if rhs.jac_is_wrapped_rhs:
-            if D.backend() != 'torch':
-                nfun_jac = utilities.JacobianWrapper(nfun, flat=True)
-                sparsity = 1.0 - D.sum(D.abs(D.to_float(nfun_jac(initial_guess))) > 0) / (self.tableau.shape[0]*self.numel)**2
-            else:
-                nfun_jac = None
-                initial_guess.requires_grad = True
-                sparsity = 1.0 - D.sum(D.jacobian(nfun(initial_guess), initial_guess) > 0) / (self.tableau.shape[0]*self.numel)**2
+        initial_guess = D.copy(self.aux)
+        
+        midpoint_guess = rhs(initial_time, initial_state, **constants)
+        self.initial_rhs   = midpoint_guess
+        midpoint_guess = 0.5 * timestep * (midpoint_guess + rhs(initial_time + 0.5 * timestep, initial_state + 0.5 * timestep * midpoint_guess, **constants))
+        
+        initial_guess = D.to_float(initial_guess + (0.5*midpoint_guess + 0.5*self.dState)[None])
+        
+        if rhs.jac_is_wrapped_rhs and D.backend() == 'torch':
+            nfun_jac = None
+            initial_guess.requires_grad = True
         else:
             nfun_jac = __nfun_jac
-            sparsity = 1.0 - D.sum(D.abs(D.to_float(nfun_jac(initial_guess))) > 0) / (self.tableau.shape[0]*self.numel)**2
             
-        try:
-            aux_root, (success, num_iter, prec) = utilities.optimizer.newtonraphson(nfun, initial_guess, jac=nfun_jac, verbose=False, tol=None, maxiter=250, sparse=sparsity >= 0.7)
-            if not success and prec > self.atol + self.rtol * D.max(D.abs(D.to_float(initial_state))):
-                raise exception_types.FailedIntegrationError("Step size too large, cannot solve system to the tolerances required: achieved = {}, desired = {}, iter = {}".format(prec, 32*D.epsilon(), num_iter))
-        except:
-            raise
+        aux_root, (success, num_iter, prec) = utilities.optimizer.newtonraphson(nfun, initial_guess, jac=nfun_jac, verbose=False, tol=None, maxiter=70)
+        if not success and prec > D.max(D.abs(self.atol + self.rtol * D.max(D.abs(D.to_float(initial_state))))):
+            raise exception_types.FailedToMeetTolerances("Step size too large, cannot solve system to the tolerances required: achieved = {}, desired = {}, iter = {}".format(prec, 32*D.epsilon(), num_iter))
             
-        self.aux = D.to_float(D.reshape(aux_root, aux_shape))
-            
-        self.dState = D.zeros_like(initial_state)
-        for fs, tbl in zip(self.final_state[0][1:], self.tableau):
-            aux_eval    = D.sum(tbl[tableau_idx_expand] * self.aux, axis=0)
-            self.dState = self.dState + timestep * fs * rhs(initial_time + tbl[0] * timestep, initial_state + timestep * aux_eval, **constants)
+        self.aux    = D.reshape(aux_root, aux_shape)
+#         if initial_state.dtype == object:
+#             new_aux = D.stack([
+#                     rhs(initial_time + tbl[0] * timestep, initial_state + timestep * D.sum(tbl[tableau_idx_expand] * self.aux, axis=0), **constants) for tbl in self.tableau
+#                 ])
+#             self.aux = new_aux
+        self.dState = timestep * D.sum(self.final_state[0][tableau_idx_expand] * self.aux, axis=0)
         self.dTime  = D.copy(timestep)
         
+        if self.final_rhs is None:
+            self.final_rhs = rhs(initial_time + self.dTime, initial_state + self.dState, **constants)
+        
         if self.adaptive:
             diff =  timestep * self.get_error_estimate(tableau_idx_expand)
             timestep, redo_step = self.update_timestep(initial_state, self.dState, diff, initial_time, timestep)
@@ -337,11 +389,39 @@ class ImplicitRungeKuttaIntegrator(ImplicitIntegrator):
 
         return timestep, (self.dTime, self.dState)
         
+    def update_timestep(self, initial_state, dState, diff, initial_time, timestep, tol=0.8):
+        err_estimate = D.max(D.abs(D.to_float(diff)))
+        relerr = D.max(D.to_float(self.atol + self.rtol * D.abs(initial_state) + self.rtol * D.abs(dState / timestep)))
+        corr = 1.0
+#         rel_niter = (self.niter0 / self.niter1)**(1.0 / self.order)
+#         if rel_niter != 0:
+#             corr = corr * rel_niter
+#         else:
+        if err_estimate != 0:
+            corr = corr * tol * (relerr / err_estimate)**(1.0 / self.order)
+        if corr != 0:
+            timestep = corr * timestep
+        if err_estimate > relerr:
+            return timestep, True
+        else:
+            return timestep, False
+        
     def get_error_estimate(self, tableau_idx_expand):
-        return D.sum((self.final_state[1] - self.final_state[0])[tableau_idx_expand] * self.aux, axis=0)
+        return D.sum(self.final_state[1][tableau_idx_expand]* self.aux, axis=0) - D.sum(self.final_state[0][tableau_idx_expand]* self.aux, axis=0)
 
     __call__ = forward
     
+    def dense_output(self):
+        return (self.initial_time + self.dTime, 
+                utilities.interpolation.CubicHermiteInterp(
+                    self.initial_time, 
+                    self.initial_time + self.dTime, 
+                    self.initial_state, 
+                    self.initial_state + self.dState,
+                    self.initial_rhs,
+                    self.final_rhs
+                ))
+    
 def generate_richardson_integrator(basis_integrator):
     """
     A function for generating an integrator that uses local Richardson Extrapolation to find the change in state ΔY over a timestep h by estimating lim ΔY as h->0.
@@ -370,14 +450,13 @@ def generate_richardson_integrator(basis_integrator):
         __implicit__   = basis_integrator.__implicit__
 
         def __init__(self, sys_dim, richardson_iter=8, **kwargs):
-            super().__init__(sys_dim, (richardson_iter, richardson_iter), True, kwargs.get('dtype'), kwargs.get('rtol', 1e-3), kwargs.get('atol', 1e-3))
+            super().__init__(sys_dim, (richardson_iter, richardson_iter), True, kwargs.get('dtype'), kwargs.get('rtol', 1e-3), kwargs.get('atol', 1e-3), kwargs.get('device', None))
             self.richardson_iter      = richardson_iter
             self.basis_integrators = [basis_integrator(sys_dim, **kwargs) for _ in range(self.richardson_iter)]
             for integrator in self.basis_integrators:
                 integrator.adaptive = False
             self.basis_order       = basis_integrator.order
             self.order             = self.basis_order + 3
-            self.device            = kwargs.get('device')
             if 'staggered_mask' in kwargs:
                 if kwargs['staggered_mask'] is None:
                     staggered_mask      = D.arange(sys_dim[0]//2, sys_dim[0], dtype=D.int64)
@@ -394,6 +473,9 @@ def generate_richardson_integrator(basis_integrator):
 
             if D.backend() == 'torch':
                 self.aux = self.aux.to(self.device)
+            
+            self.__interpolants = None
+            self.__interpolant_times = None
         
     RichardsonExtrapolatedIntegrator.__name__ = RichardsonExtrapolatedIntegrator.__qualname__ = "RichardsonExtrapolated_{}_Integrator".format(basis_integrator.__name__)
     
